import unittest
from biggu_container import Container


class FooBar:
    pass

class Foo:
    pass

class Bar:
    def __init__(self, foo: Foo):
        self.foo = foo

class Baz:
    def __init__(self, bar: Bar, foobar: FooBar):
        self.bar = bar
        self.foobar = foobar

class Dummy:
    def __init__(self, client_id, key):
        self.client_id = client_id
        self.key = key
        pass

class Defaults:
    def __init__(self, client_id, key = None, text = 'text'):
        self.client_id = client_id
        self.key = key
        self.text = text
        pass


class ContainerTest(unittest.TestCase):
    def test_ensure_key_name_as_string(self):
        container = Container()
        self.assertEqual(container.ensure_key_string(Foo), 'test.test_container.Foo')
        self.assertEqual(container.ensure_key_string("key"), 'key')
        self.assertEqual(container.ensure_key_string(1), '1')
        self.assertEqual(container.ensure_key_string(2.5), '2.5')

    def test_bind_from_lambda(self):
        container = Container()
        container.bind('key', lambda cont: 'Object')
        self.assertEqual('Object', container.make('key'))

    def test_bind_instance(self):
        container = Container()
        mock = Foo()
        container.instance('key', mock)
        self.assertEqual(mock, container.make('key'))

    def test_import_class(self):
        self.assertEqual(Container.import_class('test.test_container.Foo'), Foo)

    def test_bind_for_class_name(self):
        container = Container()
        container.bind('key', 'test.test_container.Foo')
        self.assertIsInstance(container.make('key'), Foo)

    def test_bind_for_class_name_with_dependencies(self):
        container = Container()
        container.bind('key', 'test.test_container.Baz')
        self.assertIsInstance(container.make('key'), Baz)

    def test_bind_for_class_name_with_arguments(self):
        container = Container()
        self.assertIsInstance(container.make('test.test_container.Dummy', {"client_id": "1", "key": "QWERTY1"}), Dummy)

    def test_bind_for_class_name_with_default_arguments(self):
        container = Container()
        self.assertIsInstance(container.make('test.test_container.Defaults', {"client_id": "1"}), Defaults)


    def test_singleton_instance(self):
        container = Container()
        container.singleton('singleton', 'test.test_container.Foo')
        container.singleton('singleton2', 'test.test_container.FooBar')
        self.assertEqual(container.make('singleton'), container.make('singleton'))
        self.assertEqual(container.make('singleton2'), container.make('singleton2'))

    def test_register_instance_using_class_types(self):
        container = Container()
        container.instance(Foo, Foo())
        container.instance(FooBar, FooBar())
        self.assertIsInstance(container.make('test.test_container.Foo'), Foo)
        self.assertIsInstance(container.make('test.test_container.FooBar'), FooBar)

    def test_register_singleton_using_class_types(self):
        container = Container()
        container.singleton(Foo, Foo())
        container.singleton(FooBar, FooBar())
        self.assertEqual(container.make('test.test_container.Foo'), container.make('test.test_container.Foo'))
        self.assertEqual(container.make('test.test_container.FooBar'), container.make('test.test_container.FooBar'))


if __name__ == '__main__':
    unittest.main()



from distutils.core import setup

setup(
    name = 'biggu_container',
    packages = ['biggu_container'],
    version = 'v0.0.2',
    exclude=["test"],
    description = 'IoC for python projects',
    author = 'Eduardo Salazar',
    author_email = 'eduardosalazar89@hotmail.es',
    url = 'https://github.com/esalazarv/biggu-container.git',
    download_url = 'https://github.com/esalazarv/biggu-container/archive/v0.0.2.zip',
    keywords = ['biggu_container', 'IoC', 'dependencies', 'injection', 'resolution'],
    license="MIT",
    classifiers = [
        "Programming Language :: Python :: 3.8",
    ],
    include_package_data=True,
    install_requires=[],
)


import inspect

class Container:

    def __init__(self):
        self.bindings = {}
        self.shared = {}

    @staticmethod
    def ensure_key_string(key):
        allowed = (int, str, bool, float)
        if not isinstance(key, allowed):
            key = key.__module__ + '.' + key.__name__
        return str(key)

    # Bind a resolver in biggu_container
    def bind(self, name, resolver, shared = False):
        self.bindings[name] = {
            'resolver': resolver,
            'shared': shared,
        }

    # Register an instance of class in biggu_container
    def instance(self, name, instance):
        self.shared[self.ensure_key_string(name)] = instance

    def singleton(self, name, resolver):
        self.bind(self.ensure_key_string(name), resolver, True)

    @staticmethod
    def import_class(namespace):
        print("RESOLVE CLASS NAME", namespace)
        class_data = namespace.split(".")
        submodules_list = class_data[0:-1]
        class_name = class_data[-1]

        module = __import__('.'.join(submodules_list), fromlist=[class_name])
        return getattr(module, class_name)

    # Build instances with dependencies
    def build(self, _class, arguments = None):
        _info = inspect.getfullargspec(_class.__init__)
        class_arguments = _info.args[1:]

        if not len(class_arguments):
            return _class()

        if arguments is None:
            arguments = {}

        dependencies = {}
        for key in class_arguments:
            if key in arguments:
                dependencies[key] = arguments[key]
                continue

            if key in _info.annotations and  inspect.isclass(_info.annotations[key]):
                dependencies[key] = self.build(_info.annotations[key])
            else:
                if _info.defaults is None:
                    raise Exception('Provide the value of the parameter [%s]' % key)

                last_required_index = len(class_arguments) - len(_info.defaults)
                defaults = class_arguments[last_required_index:]
                for index in range(len(defaults)):
                    name = defaults[index]
                    dependencies[name] = _info.defaults[index]


        return _class(**dependencies)

    # Make a instance using a key name in biggu_container
    def make(self, name, arguments = None):
        name = self.ensure_key_string(name)
        if name in self.shared:
            return self.shared[name]

        if name in self.bindings:
            resolver = self.bindings[name]['resolver']
            shared = self.bindings[name]['shared']
        else:
            resolver = name
            shared = False

        if callable(resolver):
            _object = resolver(self)
        else:
            if type(resolver) != str:
                _object = resolver
            else:
                _class = Container.import_class(resolver)
                _object = self.build(_class, arguments)

        if shared :
            self.instance(name, _object)

        return _object